#include "Fiedler.hh"
#include <assert.h>



// Aquestes sn les variables globals que es corresponen al hacking del fitxer chaco_sequence.c

float* jps_eigvec=0;
int* jps_eigord=0;


// Aquest sn variables externes del Chaco

extern int ECHO;
extern int OUTPUT_METRICS;
extern int OUTPUT_ASSIGN;
extern int OUTPUT_TIME;
extern int PRINT_HEADERS;
extern double SRESTOL;
extern int WARNING_EVECS;
extern int FREE_GRAPH;
extern int DEBUG_TRACE;
extern int DEBUG_PARAMS;
extern int KL_METRIC;
extern int SEQUENCE;
extern double SRESTOL;


// Aquesta s la declaraci externa de la funci interface del Chaco

extern "C" int interface (
	int       nvtxs,         /* number of vertices in full graph */
	int      *start,         /* start of edge list for each vertex */
	int      *adjacency,     /* edge list data */
	int      *vwgts,         /* weights for all vertices */
	float    *ewgts,         /* weights for all edges */
	float    *x,             /* coordinates for inertial method */
	float    *y,			 /* coordinates for inertial method */
	float    *z,             /* coordinates for inertial method */
	char     *outassignname, /* name of assignment output file */
	char     *outfilename,   /* output file name */
	short    *assignment,    /* set number of each vtx (length n) */
	int       architecture,  /* 0 => hypercube, d => d-dimensional mesh */
	int       ndims_tot,     /* total number of cube dimensions to divide */
	int       mesh_dims[3],  /* dimensions of mesh of processors */
	double   *goal,          /* desired set sizes for each set */
	int       global_method, /* global partitioning algorithm */
	int       local_method,  /* local partitioning algorithm */
	int       rqi_flag,      /* should I use RQI/Symmlq eigensolver? */
	int       vmax,          /* if so, how many vertices to coarsen down to? */
	int       ndims,         /* number of eigenvectors (2^d sets) */
	double    eigtol,        /* tolerance on eigenvectors */
	long      seed           /* for random graph mutations */
);





inline double matrix (const Graph& G, const int i, const int j) {
	if (i==j) {
		return G.deg(i);
	} else if (G.connected(i,j)) {
		return -1;
	} else {
		return 0;
	}
}


Fiedler::Fiedler (const Graph& G) 
{
	_G=&G;
	_ord=Array<int>(G.nodes());
	_vec=Array<double>(G.nodes());
	_eig=-1;
}
	

Fiedler& Fiedler::operator() ()
{
	DEBUG_TRACE=0;
	ECHO=0;
	OUTPUT_METRICS=0;
	OUTPUT_TIME=0;
	PRINT_HEADERS=0;
	WARNING_EVECS=0;
	FREE_GRAPH=0;
	DEBUG_PARAMS=0;
	SEQUENCE=1;
	SRESTOL=10.0;
	OUTPUT_ASSIGN=0;

	const int n=graph().nodes();
	const int m=graph().edges();

	int edge[2*m+1]; 
	int edge_p[n+1]; 
	jps_eigvec=new float[n]; 
	jps_eigord=new int[n]; 
	
	for (int i=0; i<n+1; i++) edge_p[i]=graph().adj(i);
	for (int i=0; i<2*m+1; i++) edge[i]=graph().edg(i)+1;

	int err=interface(
		n,		 		// int  	 nvtxs, 		 number of vertices in full graph 
		edge_p,			// int  	*start, 		 start of edge list for each vertex 
		edge,			// int  	*adjacency, 	 edge list data 
		0,				// int  	*vwgts, 		 weights for all vertices 
		0,			 	// float	*ewgts, 		 weights for all edges 
		0,			 	// float	*x, 			 coordinates for inertial method 
		0,			 	// float	*y,				 coordinates for inertial method 
		0,			 	// float	*z,				 coordinates for inertial method 
		0,			 	// char 	*outassignname,  name of assignment output file 
		0,			 	// char 	*outfilename,	 output file name 
		0,			 	// short	*assignment,	 set number of each vtx (length n) 
		0,			 	// int  	 architecture,   0 => hypercube, d => d-dimensional mesh 
		1,			 	// int  	 ndims_tot, 	 total number of cube dimensions to divide 
		0,			 	// int  	 mesh_dims[3],   dimensions of mesh of processors 
		0,		 		// double	*goal,  		 desired set sizes for each set 
		2,			 	// int  	 global_method,  global partitioning algorithm 
		2,			 	// int  	 local_method,   local partitioning algorithm 
		0,			 	// int  	 rqi_flag,  	 should I use RQI/Symmlq eigensolver? 
		1073998436,		// int  	 vmax,  		 if so, how many vertices to coarsen down to? 
		1,			 	// int  	 ndims, 		 number of eigenvectors (2^d sets) 
		0.0001,			// double	 eigtol,		 tolerance on eigenvectors 
		7654321			// long 	 seed			 for random graph mutations 
	);
	
	if (err) {
		assert(0);
		abort();
	}

	// Agafem els resultats	
	for (int i=0; i<n; i++) {
	  	_ord[jps_eigord[i]-1]=i;
		_vec[jps_eigord[i]-1]=jps_eigvec[i];
	}
	
	// Calculem el que diu en Bruce
	
	// 1. v:=u^T * M
	Array<double> v(n);	
	for (int i=0; i<n; i++) {
		double s=_vec[i]*graph().deg(i);
		for (int j=graph().adj(i); j<graph().adj(i+1); j++) {
			s-=_vec[graph().edg(j)];
		}
		v[i]=s;
	}
	
	// 2. a:= v*u
	double a=0;
	for (int i=0; i<n; i++) {
		a+=v[i]*_vec[i];
	}
	
	// 3. b:= u^T*u (normalment dona 1)
	double b=0;
	for (int i=0; i<n; i++) {
		b+=_vec[i]*_vec[i];
	}
	
	// 4. eig:=a/b;
	_eig=a/b;
	
	
	
	delete[] jps_eigord; jps_eigord=0;
	delete[] jps_eigvec; jps_eigvec=0;	
	
	return *this;
}







/*

Subject: 
	Re: 	Chaco
   	Date: 	Tue, 5 Sep 2000 10:43:58 -0600 (MDT)
	From: 	Bruce Hendrickson <bahendr@cs.sandia.gov>
	To: 	jpetit@lsi.upc.es
    CC: 	leland@sandia.gov
		
Dear Jordi -- Thanks for your message.  It's certainly
possible that there is a problem within Chaco's eigensolvers,
but they do have a number of internal checks for accuracy.

Using your notation, an eigenvector is a vector of the form
M*u = e*u, where e is an eigenvalue.  The way to determine an
eigenvalue from the eigenvector is to compute
u^T*M*u/u^T*u, where the ^T notation indicates transpose.
If I understand your message correctly, this is not the method
you are using.  I think that in exact arithmetic, your approach
would be fine.  However, in Chaco's default settings we don't
compute the eigenvector to high accuracy, which could lead to
the row-wise variation you are seeing.  What Chaco tries to do
is to computed the eigenvalue using the formula I present above.
It then looks at |M*u - e*u|.  When this norm becomes small
enough, the calculation terminates.  However, small norm-wise
errors don't guarantee small component-wise errors.  In other
words, some rows can still be poorly converged even though the
sum of them all isn't too large.

I believe that if you reduce the tolerances in Chaco you will
see your row-wise calculations get better.  We initially devised
the code for graph partitioning, in which case high-accuracy
is not needed.  For your linear arrangement work it would probably
be a good idea to use a tighter tolerance since you are using more
digits of information from the eigenvector.

If this information doesn't help, please send me your input file(s)
and the Chaco settings and inputs.  I'll try to figure out what's
going wrong.

Regards,

- Bruce


> Dear collegue,
> 
> As you know I am using Chaco and now I am facing a problem
> for which you maybe have a quick answer.
> 
> I want to compute the Fiedler vector of a graph and it
> associated eigenvalue. The eigenvector is computed by Chaco.
> The problem is that when I compute the eigenvalue from the
> graph and the vector I get *very* different values for it
> depending on "which row" I use. Some of them are negative!
> 
> To explain me better, supose M is the Laplacian matrix of
> the graph and u is the eigenvector computed by chaco. As M*u
> = e*u, I would expect that e= (sum_{i=1}^n  M[i,r]*u[i] )
> / u[r] for all row r=1..n, but this is not the case.
> 
> Please can you indicate me where I am mistaken and how to
> compute the the eigenvalue associated to the Fiedler vector?
> 
> Thank you very much in advance.
> 
> --
> Jordi Petit i Silvestre

*/
