#include <assert.h>


#include "Layout.hh"
#include "Fiedler.hh"
#include <stdlib.h>




Layout::Layout (const Graph& G) 
:	Array<int>(G.nodes()),_G(&G)
{
	set_canonical();
}


void Layout::get_lay (const string& filename)
{
	ifstream ifs(filename);
	get_lay(ifs);	
}


void Layout::put_lay (const string& filename) const
{
	ofstream ofs(filename);
	put_lay(ofs);	
}


void Layout::get_lay (istream& is)
{
	int n;	
	is >> n;
	assert(n==nodes());
	for (int i=0; i<nodes(); i++) {
		is >> phi(i);
	}
	check();
}


void Layout::put_lay (ostream& os) const
{
	os << nodes() << endl;	
	for (int i=0; i<nodes(); i++) {
		os << phi(i);
	}
	os << endl;
}


void Layout::set_canonical () 
{
	for (int i=0; i<nodes(); i++) {
		phi(i)=i;
}	}


void Layout::set_random () 
{
	set_canonical();
	permute();
}


void Layout::set_spectral () 
{
	Fiedler F(graph());
	F();
	for (int i=0; i<nodes(); i++) {
		phi(i)=F.permutation(i);
}	}



bool Layout::check (bool aborta) const
{
	const int n=nodes();
	Array<int> cnt(n);
	for (int i=0; i<n; i++) {
		cnt[i]=0;
	}
	for (int i=0; i<n; i++) {
		if (phi(i)<0 || phi(i)>=n) {
			goto ERROR;
		}
		cnt[phi(i)]++;
		if (cnt[phi(i)]>1) {
			goto ERROR;
	}	}

	return true;

	ERROR:
	if (aborta) {
		cerr << "CHECK FAILED!" << endl;
		abort();
	} else {
		return false;
}	}






int Layout::bi () const
{
	Array<int> theta(nodes());

	for (int u=0; u<nodes(); u++) {
		for (int i=graph().adj(u); i<graph().adj(u+1) && graph().edg(i)<u; i++) {
			const int v=graph().edg(i);
			if (phi(u)<phi(v)) {
	            for (int l=phi(u); l<phi(v); l++) theta[l]++;
			} else {
				for (int l=phi(v); l<phi(u); l++) theta[l]++;									 
	}	}	}

	return theta[nodes()/2];
}


int Layout::bw () const
{
	int c=0;

	for (int u=0; u<nodes(); u++) {
		for (int i=graph().adj(u); i<graph().adj(u+1) && graph().edg(i)<u; i++) {
			const int v=graph().edg(i);
			c>?=abs(phi(v)-phi(u));
	}	}	

	return c;
}


int Layout::cw () const
{
	Array<int> theta(nodes());

	for (int u=0; u<nodes(); u++) {
		for (int i=graph().adj(u); i<graph().adj(u+1) && graph().edg(i)<u; i++) {
			const int v=graph().edg(i);
			if (phi(u)<phi(v)) {
	            for (int l=phi(u); l<phi(v); l++) theta[l]++;
			} else {
				for (int l=phi(v); l<phi(u); l++) theta[l]++;									 
	}	}	}

	int c=0;	
	for (int i=0; i<nodes(); i++) c>?=theta[i];
	return c;
}



int Layout::la () const
{
	int c=0;

	for (int u=0; u<nodes(); u++) {
		for (int i=graph().adj(u); i<graph().adj(u+1) && graph().edg(i)<u; i++) {
			const int v=graph().edg(i);
			c+=abs(phi(v)-phi(u));
	}	}	

	return c;
}




int Layout::sc () const
{
	Array<int> delta(nodes());
	Array<int> gamma(nodes());

	for (int u=0; u<nodes(); u++) {
		for (int i=graph().adj(u); i<graph().adj(u+1) && graph().edg(i)<u; i++) {
			const int v=graph().edg(i);
			if (phi(u)<phi(v)) {
				if (gamma[phi(u)]<phi(v)) gamma[phi(u)]=phi(v);
			} else {
 				if (gamma[phi(v)]<phi(u)) gamma[phi(v)]=phi(u);
	}	}	}
	
	for (int i=0; i<nodes(); i++) {
		for (int j=i; j<gamma[i]; j++) {
			delta[j]++;
	}	}	

	int c=0;	
	for (int i=0; i<nodes(); i++) {
		c+=delta[i];
	}
	return c;
}


int Layout::vs () const
{
	Array<int> delta(nodes());
	Array<int> gamma(nodes());

	for (int u=0; u<nodes(); u++) {
		for (int i=graph().adj(u); i<graph().adj(u+1) && graph().edg(i)<u; i++) {
			const int v=graph().edg(i);
			if (phi(u)<phi(v)) {
				if (gamma[phi(u)]<phi(v)) gamma[phi(u)]=phi(v);
			} else {
 				if (gamma[phi(v)]<phi(u)) gamma[phi(v)]=phi(u);
	}	}	}
	
	for (int i=0; i<nodes(); i++) {
		for (int j=i; j<gamma[i]; j++) {
			delta[j]++;
	}	}	

	int c=0;	
	for (int i=0; i<nodes(); i++) {
		c>?=delta[i];
	}
	return c;
}



void Layout::all (int& bi, int& bw, int& cw, int& la, int& sc, int& vs) const
{
	Array<int> delta(nodes());
	Array<int> gamma(nodes());
	Array<int> theta(nodes());
	Array<int> lambda(2*edges());

	for (int i=0; i<nodes(); i++) {
		theta[i]=0;
		delta[i]=0;
		gamma[i]=-1;
	}
	for (int i=0; i<2*edges(); i++) {
		lambda[i]=-1;
	}

	for (int u=0; u<nodes(); u++) {
		for (int i=graph().adj(u); i<graph().adj(u+1) && graph().edg(i)<u; i++) {
			const int v=graph().edg(i);
			if (phi(u)<phi(v)) {
				lambda[i]=phi(v)-phi(u);
				if (gamma[phi(u)]<phi(v)) gamma[phi(u)]=phi(v);
	            for (int l=phi(u); l<phi(v); l++) theta[l]++;
			} else {
				lambda[i]=phi(u)-phi(v);
 				if (gamma[phi(v)]<phi(u)) gamma[phi(v)]=phi(u);
				for (int l=phi(v); l<phi(u); l++) theta[l]++;									 
	}	}	}
	
	for (int i=0; i<nodes(); i++) {
		for (int j=i; j<gamma[i]; j++) {
			delta[j]++;
	}	}

	bi=bw=cw=la=sc=vs=0;
	bi=theta[nodes()/2];
	for (int i=0; i<nodes(); i++) {
		cw>?=theta[i];
		la+=theta[i];
		vs>?=delta[i];
		sc+=delta[i];
	}
	for (int i=0; i<2*edges(); i++) {
		bw>?=lambda[i];
}	}	



string Layout::all () const
{
	int bi,bw,cw,la,sc,vs;
	
	all(bi,bw,cw,la,sc,vs);
	return string("%i %i %i %i %i %i",bi,bw,cw,la,sc,vs);
}

