#include "LowerBounds.hh"
#include "Fiedler.hh"
#include "Array2.hh"
#include <iostream.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <LEDA/set.h>



const int maxMesh = 25;

int LowerBounds_LA::gomory_hu (const Graph& G)
{
	UGRAPH<int,int> T;
	G.gomory_hu_tree(T);
	int lb=0;
	edge e;
	forall_edges (e,T) {
		lb+=T[e];
	}
	return lb;
}



int LowerBounds_LA::juvan_mohar (const Graph& G)
{
	Fiedler F(G);
	F();
	const double l2 = F.eigenvalue();
	const double n = G.nodes();
	const double lb = floor( l2 * (double(n)*double(n)-1.0) / 6.0 );
	return int(lb);
}


int LowerBounds_LA::edges_method (const Graph& G) 
{
	int lb = 0;
	const int n = G.nodes();
	const int m = G.edges();
	int i = 1; 
	int a = 0; 
	while (a+n-i <= m) {
		a += n-i;
		lb += i*(n-i);
		i ++;
	}
	lb += i*(m-a);
	return lb;
}



int LowerBounds_LA::degree_method (const Graph& G) 
{
	const int n = G.nodes();
	int lb = 0;
	for (int u=0; u<n; u++) {
		const int d = G.deg(u);
		lb += (d%2==0) ? d*d/4+d/2 : (d*d+2*d+1)/4;
	}
	lb /= 2;
	return lb;
}


int LowerBounds_LA::path_method (const Graph& G)
{
	const double n = G.nodes();
	const double m = G.edges();
	const double x = floor( (2*n-1-sqrt((2*n-1)*(2*n-1)-8*m))/2.0 );
	const double lb = x*(x+1)*(3*n-2*x-1)/6;
	return int(lb);
}





// ************************************************************************
// Mesh Method
// ************************************************************************


void print (const char* s, set<int> S) {
	int i;
	cerr << s << " = {" << endl;
	forall(i,S) cerr << i << " ";
	cerr << "}" << endl;
}




inline void admissibles1 (set<int>& A, const Graph& G, const int x, const int y, const set<int>& avail, const set<int>& used, const Array2<int>& map)
{
	if (x==1 && y==1) {
		A=avail;
	} else if (y==1) {
		const int u=map(x-1,y); 
		for (int i=G.adj(u); i<G.adj(u+1); i++) {
			const int v=G.edg(i);
			if (!used.member(v) && avail.member(v)) {
				A.insert(v);		
		}	}
	} else if (x==1) {
		const int u=map(x,y-1); 
		for (int i=G.adj(u); i<G.adj(u+1); i++) {
			const int v=G.edg(i);
			if (!used.member(v) && avail.member(v)) {
				A.insert(v);		
		}	}
	} else {
		set<int> C;
		const int u=map(x,y-1);
		const int v=map(x-1,y);
		for (int i=G.adj(u); i<G.adj(u+1); i++) {
			const int w=G.edg(i);
			if (!used.member(w) && avail.member(v)) {
				C.insert(w);
		}	}
		for (int i=G.adj(v); i<G.adj(v+1); i++) {
			const int w=G.edg(i);
			if (C.member(w)) {
				A.insert(w);
}	}	}	}






void backtrack1 (const Graph& G, 
	const int i, set<int>& avail, const int sze, 
	Array2<int>& map, 
	const Array<int>& getX, const Array<int>& getY, const Array<int>& prc, 
	bool& found, set<int>& used
)
{		
	if (i>sze*sze) {
		found=true;
	} else {	
		const int x=getX[i];
		const int y=getY[i];
		set<int> A;
		admissibles1(A,G,x,y,avail,used,map);

		int u;	
		forall (u,A) {
			if (!found) {
				if (i==1) cerr << "." << flush;
				used.insert(u);
				map(x,y)=u;
				backtrack1(G,i+1,avail,sze,map,getX,getY,prc,found,used);
				if (!found) {
					map(x,y)=-1;
					used.del(u);					
}	}	}	}	}






void find_fixed_mesh (const Graph& G, set<int> avail, const int sze, bool& found, set<int>& used)
{
	used.clear();
	found=false;
	
	Array2<int>	map(1,sze,1,sze);
	Array<int> 	getX(1,sze*sze);
	Array<int> 	getY(1,sze*sze);
	Array<int> 	prc(1,sze*sze+1);
	
	for (int i=1; i<=sze; i++) for (int j=1; j<=sze; j++) map(i,j)=-1;	
	int c=1;
	int n=1;
	while (c<=sze*sze) {
		for (int i=1; i<n; i++) {
			getX[c]=n; getY[c]=i; prc[c]=n-1; c++;
		}
		for (int i=1; i<=n; i++) {
			getX[c]=i; getY[c]=n; prc[c]=n-1; c++;
		}
		n++;
	}
	prc[sze*sze+1]=sze;	

	backtrack1(G,1,avail,sze,map,getX,getY,prc,found,used);
	cerr << endl;
}






inline void admissibles2 (set<int>& A, const Graph& G, const int x, const int y, const set<int>& used, const Array2<int>& map, const int start, const int sze)
{
	if (x==1 && y==1) {
		for (int u=0; u<G.nodes()-sze*sze+1; u++) A.insert(u);
	} else if (y==1) {
		const int u=map(x-1,y); 
		for (int i=G.adj(u); i<G.adj(u+1); i++) {
			const int v=G.edg(i);
			if (!used.member(v)) {
				A.insert(v);		
		}	}
	} else if (x==1) {
		const int u=map(x,y-1); 
		for (int i=G.adj(u); i<G.adj(u+1); i++) {
			const int v=G.edg(i);
			if (!used.member(v)) {
				A.insert(v);		
		}	}
	} else {
		set<int> C;
		const int u=map(x,y-1);
		const int v=map(x-1,y);
		for (int i=G.adj(u); i<G.adj(u+1); i++) {
			const int w=G.edg(i);
			if (!used.member(w)) {
				C.insert(w);
		}	}
		for (int i=G.adj(v); i<G.adj(v+1); i++) {
			const int w=G.edg(i);
			if (C.member(w)) {
				A.insert(w);
		}	}
	}
}





void backtrack2 (const Graph& G, 
	const int i, const int sze, 
	Array2<int>& map, 
	const Array<int>& getX, const Array<int>& getY, const Array<int>& prc, 
	bool& found, set<int>& used, const int start
)
{		
	if (i>sze*sze) {
		found=true;
	} else {	
		const int x=getX[i];
		const int y=getY[i];
		set<int> A;
		admissibles2(A,G,x,y,used,map,start,sze);

		int u;	
		forall (u,A) {
			if (!found) {
				if (i==1) {
					cerr << "." << flush;
				}
				used.insert(u);
				map(x,y)=u;
				backtrack2(G,i+1,sze,map,getX,getY,prc,found,used,start);
				if (!found) {
					map(x,y)=-1;
					used.del(u);					
}	}	}	}	}






void try_find_mesh(const Graph& G, const int sze, set<int>& used, bool& fnd, int& start)
{
	used.clear();
	fnd=false;
	
	const int s=int(ceil(sqrt(G.nodes())));
	Array2<int>	map(1,s,1,s);
	Array<int> 	getX(1,s*s);
	Array<int> 	getY(1,s*s);
	Array<int> 	prc(1,s*s+1);
	
	for (int i=1; i<=sze; i++) for (int j=1; j<=sze; j++) map(i,j)=-1;	
	int c=1;
	int n=1;
	while (c<=sze*sze) {
		for (int i=1; i<n; i++) {
			getX[c]=n; getY[c]=i; prc[c]=n-1; c++;
		}
		for (int i=1; i<=n; i++) {
			getX[c]=i; getY[c]=n; prc[c]=n-1; c++;
		}
		n++;
	}
	prc[sze*sze+1]=sze;	

	backtrack2(G,1,sze,map,getX,getY,prc,fnd,used,start);
	cerr << endl;
}



void find_max_mesh (const Graph& G, int& sze, set<int>& used)
{
	int sze2;
	set<int> used2;
	bool fnd;
	int start;
		
	cerr << "Looking for maximal mesh" << endl;
	start=0;
	sze2=1;
	try_find_mesh(G,sze2,used2,fnd,start);
	while (fnd) {
		sze=sze2; 
		used=used2;
		cerr << "Found mesh of size " << sze << endl;
		sze2++;
		if (sze2==maxMesh) break;
		try_find_mesh(G,sze2,used2,fnd,start);
	}
	cerr << "Maximal mesh has size " << sze << endl;
}


int cost (int n) {
	return int(ceil( (4-sqrt(2))/3.0*n*n*n ));
}


void write_coloring (ofstream& coloring, int& nb, const set<int>& used) {
	int u;
	forall (u,used) {
		coloring << u << "\t" << nb << endl;
	}
	nb++;	
}


int LowerBounds_LA::mesh_method (const Graph& G) 
{
	int sze;
	set<int> used;
	set<int> avail;
	bool found;
	int lb;
	int edges;
	ofstream coloring("/tmp/coloring.dat");
	int nb;
	
	coloring << G.nodes() << endl;
	
	lb=0;
	nb=0;
	edges=0;
	for (int u=0; u<G.nodes(); u++) avail.insert(u);		
	
	find_max_mesh(G, sze,used);	
	lb=cost(sze);
	edges=2*sze*(sze-1);	

	write_coloring(coloring,nb,used);	

	avail-=used;	
		
	while (sze>1) {		
		find_fixed_mesh(G,avail,sze, found,used);		
		if (found) {
			write_coloring(coloring,nb,used);	
			lb+=cost(sze);			
			edges+=2*sze*(sze-1);
			avail-=used;		
			cerr << "Found mesh of size " << sze << endl;
			cerr << "Remaining nodes: " << avail.size() << endl;
		} else {
			cerr << "No mesh of size " << sze << endl;
			sze--;
		}				
	}		
	

	// We apply now the edges method to the remaining edges
	const int n = G.nodes();
	const int m = edges;
	int i = 1; 
	int a = 0; 
	while (a+n-i <= m) {
		a += n-i;
		lb += i*(n-i);
		i ++;
	}
	lb += i*(m-a);

	return lb;
}

